#ifndef __diclib_inner_h_included__
#define __diclib_inner_h_included__


/* file_dic.h */
int anthy_init_file_dic(void);
void anthy_quit_file_dic(void);


/* xchar.c */
void anthy_init_xchar_tab(void);

#endif
